import multiprocessing as mp
import struct
import threading
import time
from threading import Thread
from typing import List

import rclpy
import serial
from rclpy.client import Client
from rclpy.context import Context
from rclpy.node import Node
from rclpy.parameter import Parameter
from std_msgs.msg import Float32MultiArray

start_time = time.time() * 1000


def read_message_from_serial_stream_and_publish(stream: serial.Serial, speed_publisher):
    """
    Read a message from the given serial stream.

    :param stream: The serial stream.
    :return: The message.
    """
    while stream.is_open:
        # Read until we find the start of a message.
        while stream.read(1) != b"\xAA":
            pass

        # Read the rest of the message.
        frame_type = stream.read(1)
        if frame_type.hex() == "00":
            message_type = "measurement"
        elif frame_type.hex() == "20":
            message_type = "command"
        elif frame_type.hex() == "40":
            message_type = "response"
        else:
            # print("Invalid frame type: %s" % frame_type)
            continue

        frame_length = stream.read(1)
        frame_component = stream.read(1)
        if frame_component.hex() == "00":
            component = "default"
        elif frame_component.hex() == "10":
            component = "left_motor"
        elif frame_component.hex() == "20":
            component = "right_motor"
        elif frame_component.hex() == "30":
            component = "both_motor"
        else:
            # print("Invalid frame component: %s" % frame_component.hex())
            continue

        timestamp = stream.read(4)
        payload = stream.read(int.from_bytes(frame_length, byteorder="big"))

        checksum = stream.read(2)

        # Check the checksum.
        if sum(payload) != int.from_bytes(checksum, byteorder="big"):
            # print("Invalid checksum")
            # print(message_type,component,timestamp,payload,checksum)
            continue

        # Handle message contents.
        if message_type == "measurement":
            if component == "left_motor":
                track_speed_left = struct.unpack(">f", payload[0:4])[0]
            elif component == "right_motor":
                track_speed_right = struct.unpack(">f", payload[0:4])[0]
            elif component == "both_motor":
                try:
                    track_speed_right = struct.unpack(">f", payload[0:4])[0]
                    track_speed_left = struct.unpack(">f", payload[4:8])[0]
                except Exception as e:
                    # print(e)
                    # print("Invalid payload: %s" % payload.hex())
                    continue
                msg = Float32MultiArray()
                msg.data = [track_speed_left, track_speed_right]
                # msg.header.stamp = node.get_clock().now().to_msg()
                speed_publisher.publish(msg)
        elif message_type == "command":
            if payload == b"\x02":
                print("RESET_TIMESTAMP")
                global start_time
                start_time = time.time() * 1000


def create_uart_message(
    message_type="default", component="default", payload=b"", timestamp=0
):
    """
    Create a UART message.

    :param message_type: The type of the message.
    :param component: The component of the message.
    :param payload: The payload of the message.
    :param timestamp: The timestamp of the message.
    :return: The UART message.
    """
    # Create the header.
    message_type_bytes = None
    if message_type == "measurement":
        message_type_bytes = b"\x00"
    elif message_type == "command":
        message_type_bytes = b"\x20"
    elif message_type == "response":
        message_type_bytes = b"\x40"
    else:
        raise ValueError("Invalid message type: %s" % message_type)

    component_bytes = None
    if component == "default":
        component_bytes = b"\x00"
    elif component == "left_motor":
        component_bytes = b"\x10"
    elif component == "right_motor":
        component_bytes = b"\x20"
    elif component == "both_motor":
        component_bytes = b"\x30"
    else:
        raise ValueError("Invalid component: %s" % component)

    # Create the payload.
    payload = bytearray(payload)

    # Create the timestamp.
    timestamp = bytearray(timestamp.to_bytes(4, byteorder="big"))

    # Create the checksum.
    checksum = bytearray(sum(payload).to_bytes(2, byteorder="little"))

    # Create the message.
    message = bytearray()
    message.append(0xAA)
    message.extend(message_type_bytes)
    message.extend(len(payload).to_bytes(1, byteorder="big"))
    message.extend(component_bytes)
    message.extend(timestamp)
    message.extend(payload)
    message.extend(checksum)

    # print(''.join([ "%02x" % b for b in message ]))

    return message


class ROS2UARTNode(Node):
    def __init__(
        self,
        node_name: str,
        ser: serial.Serial,
        *,
        context: Context = None,
        cli_args: List[str] = None,
        namespace: str = None,
        use_global_arguments: bool = True,
        enable_rosout: bool = True,
        start_parameter_services: bool = True,
        parameter_overrides: List[Parameter] = None,
        allow_undeclared_parameters: bool = False,
        automatically_declare_parameters_from_overrides: bool = False
    ) -> None:
        super().__init__(
            node_name,
            context=context,
            cli_args=cli_args,
            namespace=namespace,
            use_global_arguments=use_global_arguments,
            enable_rosout=enable_rosout,
            start_parameter_services=start_parameter_services,
            parameter_overrides=parameter_overrides,
            allow_undeclared_parameters=allow_undeclared_parameters,
            automatically_declare_parameters_from_overrides=automatically_declare_parameters_from_overrides,
        )
        self.create_subscription(Float32MultiArray, "/track_speeds", self.callback, 1)
        self.left_speed = 0
        self.right_speed = 0

        self.ser = ser

    def callback(self, msg: Float32MultiArray) -> None:
        print(msg.data)
        assert len(msg.data) == 2
        self.left_speed, self.right_speed = msg.data
        self.left_speed /= 1
        self.right_speed /= 1
        left_byte = struct.pack(">f", self.left_speed)
        right_byte = struct.pack(">f", self.right_speed)

        self.ser.write(
            create_uart_message(
                message_type="command",
                component="both_motor",
                payload=b"\x01" + left_byte + right_byte,
                timestamp=0,
            )
        )
        self.ser.flush()


class UART2ROSNode(Node):
    def __init__(
        self,
        node_name: str,
        ser: serial.Serial,
        *,
        context: Context = None,
        cli_args: List[str] = None,
        namespace: str = None,
        use_global_arguments: bool = True,
        enable_rosout: bool = True,
        start_parameter_services: bool = True,
        parameter_overrides: List[Parameter] = None,
        allow_undeclared_parameters: bool = False,
        automatically_declare_parameters_from_overrides: bool = False
    ) -> None:
        super().__init__(
            node_name,
            context=context,
            cli_args=cli_args,
            namespace=namespace,
            use_global_arguments=use_global_arguments,
            enable_rosout=enable_rosout,
            start_parameter_services=start_parameter_services,
            parameter_overrides=parameter_overrides,
            allow_undeclared_parameters=allow_undeclared_parameters,
            automatically_declare_parameters_from_overrides=automatically_declare_parameters_from_overrides,
        )
        self.speed_publisher = self.create_publisher(Float32MultiArray, "topic", 10)
        self.ser = ser

        read_message_from_serial_stream_and_publish()

    def destroy_node(self) -> bool:
        self.process.join()
        return super().destroy_node()


def main(args=None) -> None:
    with serial.Serial("/dev/ttyUSB1", 115200, timeout=0.001) as ser:
        rclpy.init(args=args)
        ros2uart = ROS2UARTNode(node_name="ros2uart", ser=ser)
        # uart2ros = UART2ROSNode(node_name='uart2ros', ser=ser)

        executor = rclpy.executors.MultiThreadedExecutor()
        # executor.add_node(uart2ros)
        executor.add_node(ros2uart)
        executor_thread = threading.Thread(target=executor.spin, daemon=True)
        executor_thread.start()

        # loop while not interrupted
        curr_time = time.time()
        rate = ros2uart.create_rate(2)
        try:
            while rclpy.ok():
                print("Help me body, you are my only hope")
                rate.sleep()
        except KeyboardInterrupt:
            pass
        rclpy.shutdown()
        executor_thread.join()


if __name__ == "__main__":
    main()
