import math
from typing import List

import numpy as np
import rclpy
from rclpy.context import Context
from rclpy.node import Node
from rclpy.parameter import Parameter
from sensor_msgs.msg import Joy
from std_msgs.msg import Float32MultiArray


def calc_left_power(angle: float) -> float:
    angle_deg = math.degrees(angle)
    if -180 <= angle_deg < -90:
        return 2 / 90 * (angle_deg + 180) - 1
    elif -90 <= angle_deg < 0:
        return 1
    elif 0 <= angle_deg < 90:
        return -2 / 90 * angle_deg + 1
    elif 90 <= angle_deg <= 180:
        return -1


def calc_right_power(angle: float) -> float:
    angle_deg = math.degrees(angle)
    if -180 <= angle_deg < -90:
        return -1
    elif -90 <= angle_deg < 0:
        return 2 / 90 * (angle_deg + 90) - 1
    elif 0 <= angle_deg < 90:
        return 1
    elif 90 <= angle_deg <= 180:
        return -2 / 90 * (angle_deg - 90) + 1


def calc_power_from_euclidean_joystick(x, y):
    x, y = y, -x  # rotate 90 so that up is 0 degress
    r = np.linalg.norm([x, y]) / np.sqrt(2)
    angle = np.arctan2(y, x)

    return calc_left_power(angle) * r, calc_right_power(angle) * r


class JoystickSpeedDistributorNode(Node):
    def __init__(
        self,
        node_name: str,
        *,
        context: Context = None,
        cli_args: List[str] = None,
        namespace: str = None,
        use_global_arguments: bool = True,
        enable_rosout: bool = True,
        start_parameter_services: bool = True,
        parameter_overrides: List[Parameter] = None,
        allow_undeclared_parameters: bool = False,
        automatically_declare_parameters_from_overrides: bool = False
    ) -> None:
        super().__init__(
            node_name,
            context=context,
            cli_args=cli_args,
            namespace=namespace,
            use_global_arguments=use_global_arguments,
            enable_rosout=enable_rosout,
            start_parameter_services=start_parameter_services,
            parameter_overrides=parameter_overrides,
            allow_undeclared_parameters=allow_undeclared_parameters,
            automatically_declare_parameters_from_overrides=automatically_declare_parameters_from_overrides,
        )
        self.l = 0.1

        self.create_subscription(Joy, "/joy", self.joy_callback, 1)

        self.pub = self.create_publisher(Float32MultiArray, "/track_speeds", 1)

    def joy_callback(self, joy_message):
        x, y = joy_message.axes[:2]
        y = -y

        powers = calc_power_from_euclidean_joystick(x, y)

        msg = Float32MultiArray()

        msg.data = list(map(float, powers))

        self.pub.publish(msg)


def main(args=None) -> None:
    rclpy.init(args=args)
    node = JoystickSpeedDistributorNode("speed_distributor")
    rclpy.spin(node)
    rclpy.shutdown()


if __name__ == "__main__":
    main()
