#include "communication/UARTCom.hpp"
#include "communication/UCommands.hpp"
#include "communication/UValue.hpp"
#include "misc/Utils.hpp"
#include "constants.h"

uint8_t UARTCom::data[MAX_PACKET_LENGTH];
uint8_t currentParseIndex;
int payloadLength;
uint32_t currentTimestamp = 0;
extern MotorWheel wheelLeft;
extern MotorWheel wheelRight;

/**
 * @brief extracts frametype from data package
 * 
 * @param d data package
 * @return U_FrameType 
 */
U_FrameType parseFrameType (uint8_t* d)
{
	uint8_t b = d[FRAME_TYPE_POS];
	return (U_FrameType) MID(b, FRAME_TYPE_ENDBIT, FRAME_TYPE_STARTBIT + 1);
}

/**
 * @brief translates frametype to corresponding byte value
 * 
 * @param ft frametype
 * @return uint8_t 
 */
uint8_t fromFrameType (U_FrameType ft)
{
	switch (ft) {
		case VALUE: return FT_VALUE_BV;
		case REQUEST: return FT_REQUEST_BV;
		case RESPONSE: return FT_RESPONSE_BV;
		default: return 1;
	}
}

/**
 * @brief extract component information 
 * 
 * @param d data package
 * @return U_Component 
 */
U_Component parseComponent (uint8_t* d)
{
	uint8_t b = d[COMPONENT_POS];
	return (U_Component) MID(b, COMPONENT_ENDBIT, COMPONENT_STARTBIT + 1);
}

/**
 * @brief translates component into byte value
 * 
 * @param cp component
 * @return uint8_t 
 */
uint8_t fromComponent (U_Component cp)
{
	switch (cp) {
		case DEFAULT_COMPONENT: return CP_DEFAULT_BV;
		case LEFT_MOTOR: return CP_LEFT_M_BV;
		case RIGHT_MOTOR: return CP_RIGHT_M_BV;
		case ALL_MOTORS: return CP_ALL_M_BV;
		default: return 1;
	}
}

/**
 * @brief extracts request information from data package
 * 
 * @param d data package
 * @return U_Request 
 */
U_Request parseRequest (uint8_t* d)
{
	return (U_Request) d[COMMAND_POS];
}

/**
 * @brief extracts timestamp from data package
 * 
 * @param d data package
 * @return uint32_t 
 */
uint32_t parseTimestamp (uint8_t* d)
{
	return Utils::bytesToInt(d[TIMESTAMP_POS],
							 d[TIMESTAMP_POS + 1],
							 d[TIMESTAMP_POS + 2],
							 d[TIMESTAMP_POS + 3]);
}

/**
 * @brief computes checksum from data package
 * 
 * @param d data package
 * @param pl payload length
 * @return uint16_t 
 */
uint16_t computeChecksum (uint8_t* d, uint8_t pl)
{
	uint16_t r = 0;
	for (int i = PAYLOAD_POS; i < PAYLOAD_POS + pl; i++) {
		r += (uint8_t) d[i];
	}
	return r;
}

/**
 * @brief checks if checksum inside package and computed checksum from package data are the same
 * 
 * @param d data package
 * @return int 
 */
int correctChecksum (uint8_t* d)
{
	uint8_t payloadLength = d[PAYLOAD_LENGTH_POS];
	uint16_t checksumRead = d[PACKAGE_LENGTH_NO_P + payloadLength - 1] | d[PACKAGE_LENGTH_NO_P + payloadLength - 2] << 8;
	return checksumRead == computeChecksum(d, payloadLength);
}

/**
 * @brief debugging function, prints out package data
 * 
 * @param d data package
 * @param pl payload length
 */
void showPackage (uint8_t* d, int pl)
{
	Serial.println("Package to be handled:");
	for (int i = 0; i < PACKAGE_LENGTH_NO_P + pl; i++) {
		char tmp[3];
		sprintf(tmp, "%02X", d[i]);
		Serial.print(tmp);
	}
	Serial.println();
}

/**
 * @brief erases all package data and sets the array to 0
 * 
 * @param d data package
 */
void clearPackage (uint8_t* d)
{
	for (int i = 0; i < MAX_PACKET_LENGTH; i++) {
		d[i] = 0;
	}
	currentParseIndex = 0;
	payloadLength = -1;
}

/**
 * @brief executes commands according to incoming request
 * 
 * @param d data package
 */
void handleRequest (uint8_t* d)
{
	switch (parseRequest(d)) {
		case DRIVE_MOTOR: {
			UCommands::setSetpoint(d, parseComponent(d));
			break;
		}
		case RESET_TIMESTAMP: {
			currentTimestamp = 0;
			break;
		}
		case DRIVE_ENCODER: {
			UValue::sendBothEncoderValues(-wheelLeft.getGearedSpeedRPM(), wheelRight.getGearedSpeedRPM());
			break;
		}
		case PID_PARAMETER: {
			UCommands::setPIDParameter(d, parseComponent(d));
			break;
		}
		default: {
			Serial.println("UNKNOWN REQUEST");
			break;
		}
	}
}

/**
 * @brief package handler, checks for errors first, then proceeds to
 * extract information from the data package
 * 
 * @param d 
 * @param pl 
 */
void handlePackage (uint8_t* d, int pl)
{
	if (d[START_BYTE_POS] != START_BYTE || !correctChecksum(d)) {
		Serial.println("INCORRECT START BYTE OR CHECKSUM");
		return;
	}
	#ifdef DCAITI_DEBUG
	showPackage(d, pl);
	#endif
	/*
	if (parseTimestamp(d) <= currentTimestamp) {
		if (parseFrameType(d) == REQUEST && parseRequest(d) == RESET_TIMESTAMP) {
			currentTimestamp = 0;
		}
		else {
			return;
		}
	}
	else if (parseTimestamp(d) - currentTimestamp < 10000) {
		currentTimestamp = parseTimestamp(d);
	} 
	else return;*/
	switch (parseFrameType(d)) {
		case VALUE: {
			// TODO, currently not used
			break;
		}
		case REQUEST: {
			handleRequest(d);
			break;
		}
		case RESPONSE: {
			// TODO, currently not used
			break;
		}
		default: {
			break;
		}
	}
}

void UARTCom::init (uint8_t* d)
{
	Serial.begin(UART_BAUD_RATE);
	clearPackage(d);
}

void UARTCom::read (uint8_t* d)
{
	uint8_t byteRead = 0;
	while (Serial.available()) {
		byteRead = Serial.read();

		#ifdef DCAITI_DEBUG
		Serial.print("INCOMING: ");

		char tmp[3];
		sprintf(tmp, "%02X", byteRead);
		Serial.println(tmp);
		#endif

		switch (currentParseIndex) {
			case START_BYTE_POS: {
				if (byteRead != START_BYTE) {
					clearPackage(d);
					return;
				}
				d[currentParseIndex] = START_BYTE;
				break;
			}
			case FRAME_TYPE_POS: {
				d[currentParseIndex] = byteRead;
				break;
			}
			case PAYLOAD_LENGTH_POS: {
				payloadLength = byteRead;
				d[currentParseIndex] = payloadLength;
				break;
			}
			default: {
				d[currentParseIndex] = byteRead;
				break;
			}
		}
		if (payloadLength != -1 && currentParseIndex == PACKAGE_LENGTH_NO_P + payloadLength - 1) {
			handlePackage(d, payloadLength);
			clearPackage(d);
		}
		else currentParseIndex++;
	}
}

void UARTCom::send (U_FrameType ft, uint8_t pll, U_Component cp, uint8_t* pl)
{
	uint8_t toSend[PACKAGE_LENGTH_NO_P + pll];
	for (int i = 0; i < PACKAGE_LENGTH_NO_P + pll; i++) toSend[i] = 0;

	toSend[START_BYTE_POS] = START_BYTE;
	toSend[FRAME_TYPE_POS] = fromFrameType(ft);
	toSend[PAYLOAD_LENGTH_POS] = pll;
	toSend[COMPONENT_POS] = fromComponent(cp);
	
	union {
		uint32_t t;
		uint8_t b[4];
	} u;
	u.t = (uint32_t) millis();

	toSend[TIMESTAMP_POS] = u.b[3];
	toSend[TIMESTAMP_POS + 1] = u.b[2];
	toSend[TIMESTAMP_POS + 2] = u.b[1];
	toSend[TIMESTAMP_POS + 3] = u.b[0];

	uint16_t checksum = 0;

	for (int i = 0; i < pll; i++) {
		toSend[PAYLOAD_POS + i] = pl[i];
		checksum += pl[i];
	}

	union {
		uint16_t c;
		uint8_t b[2];
	} ch;

	ch.c = checksum;

	toSend[PAYLOAD_POS + pll] = ch.b[1];
	toSend[PAYLOAD_POS + pll + 1] = ch.b[0];

	Serial.write(toSend, sizeof(toSend));
}