#include "diffdrive_arduino/arduino_comms.h"
// #include <ros/console.h>
#include <rclcpp/rclcpp.hpp>
#include <sstream>
#include <cstdlib>
#include <arpa/inet.h>
#include <vector>


void ArduinoComms::setup(const std::string &serial_device, int32_t baud_rate, int32_t timeout_ms)
{  
    serial_conn_.setPort(serial_device);
    serial_conn_.setBaudrate(baud_rate);
    serial::Timeout tt = serial::Timeout::simpleTimeout(timeout_ms);
    serial_conn_.setTimeout(tt); // This should be inline except setTimeout takes a reference and so needs a variable
    serial_conn_.open();

}

void ArduinoComms::sendEmptyMsg()
{
}

void ArduinoComms::readEncoderValues(double &left_speed, double &right_speed)
{   
    std::vector<uint8_t> payload = {DRIVE_ENCODER};
    std::vector<uint8_t> msg = buildMsg(REQUEST, BOTH_MOTORS, 0, payload);
    sendMsg(msg);
    readSpeed(left_speed, right_speed);
}

void ArduinoComms::setMotorValues(float left_speed, float right_speed)
{
    
    std::vector<uint8_t> payload = {DRIVE_MOTOR};
    auto bytes_left_speed = marshalling(left_speed);
    payload.insert(payload.end(), bytes_left_speed.begin(), bytes_left_speed.end());
    auto bytes_right_speed = marshalling(right_speed);
    payload.insert(payload.end(), bytes_right_speed.begin(), bytes_right_speed.end());

    std::vector<uint8_t> msg = buildMsg(REQUEST, BOTH_MOTORS, 0, payload);

    sendMsg(msg);
}

void ArduinoComms::setPidValues(float k_p, float k_d, float k_i, float k_o)
{
    std::stringstream ss;
    ss << "u " << k_p << ":" << k_d << ":" << k_i << ":" << k_o << "\r";
}

void ArduinoComms::readSpeed(double &left_speed, double &right_speed)
{
    std::vector<uint8_t> buffer;
    int read_bytes = 0;
    do{
        buffer.clear();
        read_bytes = serial_conn_.read(buffer, 1);
    }
    while(read_bytes == 0 || buffer.at(0) != 0xAA);
    
    // read frame and extract two big-endian floats from payload
    safeRead(buffer, 1);
    uint8_t frame_type = buffer.at(0);
    safeRead(buffer, 1);
    uint8_t payload_size = buffer.at(0);
    safeRead(buffer, 1);
    uint8_t frame_component = buffer.at(0);
    safeRead(buffer, 4);
    uint32_t time = *reinterpret_cast<double*>(&buffer[0]);

    safeRead(buffer, payload_size);
    std::vector<uint8_t> payload(buffer.begin(), buffer.end());
        // read checksum
    safeRead(buffer, 2);
    uint16_t checksum = buffer.at(0);
    checksum = checksum << 8;
    checksum = checksum | buffer.at(1);

    // check checksum of payload
    uint16_t payload_checksum = 0;
    for (auto it = payload.begin(); it != payload.end(); ++it)
    {
        payload_checksum += *it;
    }
    if (payload_checksum != checksum)
    {
        RCLCPP_ERROR(rclcpp::get_logger("rclcpp"), "Checksum error");
    }
    
    if (frame_type == VALUE && frame_component == BOTH_MOTORS)
    {
            std::vector<uint8_t> left_speed_bytes(payload.begin(), payload.begin() + 4);
            std::vector<uint8_t> right_speed_bytes(payload.begin() + 4, payload.end());
            std::reverse(left_speed_bytes.begin(), left_speed_bytes.end());
            std::reverse(right_speed_bytes.begin(), right_speed_bytes.end());
            float left_speed_float = *reinterpret_cast<float*>(&left_speed_bytes[0]);
            float right_speed_float = *reinterpret_cast<float*>(&right_speed_bytes[0]);
            left_speed = left_speed_float;
            right_speed = right_speed_float;
    }
}

void ArduinoComms::safeRead(std::vector<uint8_t> &buffer, int bytes_to_read){
    buffer.clear();
    int read_bytes = 0;
    do{
        read_bytes += serial_conn_.read(buffer, bytes_to_read - read_bytes);
    }
    while(read_bytes < bytes_to_read);
}

std::vector<uint8_t> ArduinoComms::marshalling(uint16_t number)
{
    number = htons(number);
    uint8_t array[2];
    array[0]=number & 0xff;
    array[1]=(number >> 8);
    std::vector<uint8_t> vec(array, array + sizeof array / sizeof array[0]);
    return vec;
}

std::vector<uint8_t> ArduinoComms::marshalling(uint32_t number)
{
    number = htons(number);
    uint8_t array[4];
    array[0]=number & 0xff;
    array[1]=(number >> 8) & 0xff;
    array[2]=(number >> 16) & 0xff;
    array[3]=(number >> 24) & 0xff;
    std::vector<uint8_t> vec(array, array + sizeof array / sizeof array[0]);
    return vec;
}

std::vector<uint8_t> ArduinoComms::marshalling(float number)
{
    uint32_t int_interpretation = reinterpret_cast<uint32_t &>( number );
    union {
    float f;
    uint8_t b[4];
    } u;
    u.f = number;
    std::vector<uint8_t> vec(u.b, u.b + sizeof u.b / sizeof u.b[0]);
    std::reverse(vec.begin(), vec.end());
    return vec;
}

const std::vector<uint8_t> ArduinoComms::buildMsg(uint8_t frame_type, uint8_t frame_component, long time,  std::vector<uint8_t> payload)
{
    std::vector<uint8_t> buffer;
    buffer.push_back(0xAA);
    buffer.push_back(frame_type);
    buffer.push_back(payload.size());
    buffer.push_back(frame_component);
    uint32_t time_zero = 0;
    auto byte_time = marshalling(time_zero);
    buffer.insert(buffer.end(), byte_time.begin(), byte_time.end());
    uint16_t checksum = std::accumulate(payload.begin(), payload.end(), 0);
    buffer.insert(buffer.end(), payload.begin(), payload.end());
    auto byte_checksum = marshalling(checksum);
    buffer.insert(buffer.end(), byte_checksum.begin(), byte_checksum.end());

    return buffer;
}


void ArduinoComms::sendMsg(const std::vector<uint8_t> &msg_to_send, bool print_output)
{
    serial_conn_.write(msg_to_send);
}